# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# fmt: off
from .....schema_classes import AssertionInfoClass
from .....schema_classes import AssertionResultClass
from .....schema_classes import AssertionScopeClass
from .....schema_classes import AssertionStdOperatorClass
from .....schema_classes import AssertionTypeClass
from .....schema_classes import BatchAssertionResultClass
from .....schema_classes import BatchSpecClass
from .....schema_classes import DatasetColumnAssertionClass
from .....schema_classes import DatasetColumnStdAggFuncClass
from .....schema_classes import DatasetRowsAssertionClass
from .....schema_classes import DatasetRowsStdAggFuncClass
from .....schema_classes import DatasetSchemaAssertionClass
from .....schema_classes import DatasetSchemaStdAggFuncClass


AssertionInfo = AssertionInfoClass
AssertionResult = AssertionResultClass
AssertionScope = AssertionScopeClass
AssertionStdOperator = AssertionStdOperatorClass
AssertionType = AssertionTypeClass
BatchAssertionResult = BatchAssertionResultClass
BatchSpec = BatchSpecClass
DatasetColumnAssertion = DatasetColumnAssertionClass
DatasetColumnStdAggFunc = DatasetColumnStdAggFuncClass
DatasetRowsAssertion = DatasetRowsAssertionClass
DatasetRowsStdAggFunc = DatasetRowsStdAggFuncClass
DatasetSchemaAssertion = DatasetSchemaAssertionClass
DatasetSchemaStdAggFunc = DatasetSchemaStdAggFuncClass

# fmt: on
