# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# fmt: off
from .....schema_classes import DashboardInfoClass
from .....schema_classes import EditableDashboardPropertiesClass


DashboardInfo = DashboardInfoClass
EditableDashboardProperties = EditableDashboardPropertiesClass

# fmt: on
